#include <stdio.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/wait.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/msg.h>
#include <pthread.h>
#include <assert.h>
#include <semaphore.h>
#include <fpu_control.h>

#define MQ_DEF_MSGSIZE 16
#define MQ_DEF_MAXMSG 10

extern "C" void march_interrupt_solver();
void minisat_interrupt_solver();
extern "C" int march_main(int argc, char** argv);
int minisat_main(int argc, char** argv);

pthread_t marchThread;
pthread_t minisatThread;
sem_t signalExitSemaphore;
key_t queueToMinisatKey = -1;
key_t queueToMarchKey = -1;
int queueToMinisat = -1;
int queueToMarch = -1;

static void cleanup()
{
	msgctl(queueToMinisat, IPC_RMID, NULL);
	msgctl(queueToMarch, IPC_RMID, NULL);
}

static void handleSignal(int signum)
{
	switch (signum)
	{
	case SIGINT:
	case SIGTERM:
	{
		printf("Received signal %d as %08X. Exiting...\n", signum, pthread_self());
		minisat_interrupt_solver();
		march_interrupt_solver();
		timespec joinEndTime;
		clock_gettime(CLOCK_REALTIME, &joinEndTime);
		joinEndTime.tv_sec += 1;
		sem_timedwait(&signalExitSemaphore, &joinEndTime);
		int waitResult = sem_timedwait(&signalExitSemaphore, &joinEndTime); // Wait twice, for march and minisat.
		if (waitResult == -1 && errno == ETIMEDOUT)
		{
			printf("Either or both solvers did shut down gracefully. Force quitting.\n");
		}
		pthread_cancel(minisatThread);
		pthread_cancel(marchThread);
		cleanup();
		signal(signum, SIG_DFL);
		kill(getpid(), signum);
		break;
	}
	default:
		assert(0);
	}
}

struct StartupInfo
{
	int argc;
	char** argv;
};

void* start_minisat_main(void* arg)
{
	void* result;
	StartupInfo* startupInfo = (StartupInfo*)arg;
	printf("Minisat has started.\n");
	result = (void*)minisat_main(startupInfo->argc, startupInfo->argv);
	printf("Minisat has stopped with exit code %ld.\n", (long)result);
	sem_post(&signalExitSemaphore);
	return result;
}

void* start_march_main(void* arg)
{
	void* result;
	StartupInfo* startupInfo = (StartupInfo*)arg;
	printf("March has started.\n");
	result = (void*)march_main(startupInfo->argc, startupInfo->argv);
	printf("March has stopped with exit code %ld.\n", (long)result);
	sem_post(&signalExitSemaphore);
	return result;
}

int main(int argc, char** argv)
{
	setlinebuf(stdout);
	fpu_control_t oldcw, newcw;
	_FPU_GETCW(oldcw); newcw = (oldcw & ~_FPU_EXTENDED & ~_FPU_SINGLE) | _FPU_DOUBLE; _FPU_SETCW(newcw); // For repeatability

	if (argc < 2)
	{
		printf("Usage: ccc [options] <instance file>\n");
		return 1;
	}

	const char* instanceFilename = argv[argc - 1];
	if (access(instanceFilename, F_OK) != 0)
	{
		printf("Instance cannot be found.\n");
		return 1;
	}

	queueToMinisatKey = ftok(".", 0);
	if (queueToMinisatKey == -1) {
		perror("ftok");
		return 1;
	}

	queueToMarchKey = ftok(".", 1);
	if (queueToMarchKey == -1) {
		perror("ftok");
		return 1;
	}

	queueToMinisat = msgget(queueToMinisatKey, 0600 | IPC_CREAT | IPC_EXCL);
	if (queueToMinisat == -1) {
		perror("msgget");
		return 1;
	}

	queueToMarch = msgget(queueToMarchKey, 0600 | IPC_CREAT | IPC_EXCL);
	if (queueToMarch == -1) {
		perror("msgget");
		return 1;
	}

	sem_init(&signalExitSemaphore, 0, 0);

	atexit(&cleanup);
	signal(SIGINT, handleSignal);
	signal(SIGTERM, handleSignal);

	StartupInfo startupInfoMarch = { argc, argv };
	StartupInfo startupInfoMinisat = { argc, new char*[argc] }; memcpy(startupInfoMinisat.argv, argv, sizeof(char*)*argc);
	pthread_create(&marchThread, NULL, start_march_main, (void*)&startupInfoMarch);
	pthread_create(&minisatThread, NULL, start_minisat_main, (void*)&startupInfoMinisat);
	void *marchResult, *minisatResult;
	pthread_join(marchThread, &marchResult);
	pthread_join(minisatThread, &minisatResult);

	// If minisat found a solution, return it. Otherwise, return the result found by march.
	if ((long)minisatResult == 10 || (long)minisatResult == 20)
		return (long)minisatResult;
	else
		return (long)marchResult;
}
