#include "ccc.h"
#include "common.h"
#include <stdio.h>
#include <memory.h>
#include <assert.h>
#include <stdint.h>
#include <errno.h>
#include <stdlib.h>
#include <sys/ipc.h>
#include <sys/msg.h>
#include <signal.h>
#include <unistd.h>
#include <stdint.h>
#include <math.h>
#include <time.h>

#define SHOW_DEBUG_OUTPUT 0
#define CCC_LOG_CUBES 0
#define CCC_LOG_SLOW_CUBES 0
#define CCC_LOG_SLOW_CUBES_THRESHOLD 10000000000
#define RT_ASSERT(x) { if (!(x)) { printf("ASSERTION FAILED: (%s) @ %s:%d\n", #x, __FILE__, __LINE__); exit(0); } }


#define CCC_MESSAGE_TYPE_EXIT 1 // {} No messages may be sent after sending a message of this type.
#define CCC_MESSAGE_TYPE_DECISION 2 // {id, level, literal} The lookahead solver makes a new cube by adding a decision literal, the cdcl solver should update its current cube upon receiving a message of this type.
#define CCC_MESSAGE_TYPE_UNSAT 3 // {id} The cdcl solver proved unsatisfiability of a cube.
typedef struct
{
	int64_t ipc_type;
	int32_t type;
	int32_t arg1;
	int32_t arg2;
	int32_t arg3;
} ccc_Message;

typedef struct
{
	int id;
	int decision;
	int discrepancies;
	float score;
	int isLeaf;
	int isDiscrepancy;
	int isSolvedByCdcl;
	int cutoffChildCount;
	int refutedChildDecision;
	int numberOfLeafNodesRefutedByCdcl;
	int numberOfLeafNodesRefutedByLookahead;
} ccc_Cube;

typedef struct
{
	uint_fast32_t numberOfCubesOpened;
	uint_fast32_t numberOfCutoffCubesOpened;
	uint_fast32_t numberOfLearntClauses;
	uint_fast32_t numberOfCubesDispatchedToCdcl;
	uint_fast32_t numberOfCubesProvedUnsatByCdcl; // (can include sub cubes)
	uint_fast32_t numberOfCubesProvedUnsatByCdclAndIgnored; // ignored because the result was already known. included in numberOfCubesProvedUnsatByCdcl
	uint_fast32_t numberOfInternalCubesClosed;
	uint_fast32_t numberOfLeafCubesClosedDueToCdcl;
	uint_fast32_t numberOfLeafCubesClosedDueToLookahead;
	uint_fast64_t sumOfLengthOfDispatchedCubes;
	uint_fast64_t sumOfLengthOfClosedLeafCubes;
	uint_fast32_t maxNumberOfLevelsBacktracked;
	uint_fast64_t sumOfNumberOfLevelsBacktracked;
} ccc_Stats;

static key_t queueToMinisatKey = -1;
static key_t queueToMarchKey = -1;
static int queueToMinisat = -1;
static int queueToMarch = -1;
static int nextLevel;
static int nextNodeId;
static ccc_Cube* trail = NULL; // [0] is the root node and has no (meaningful) value for .decision. [i] corresponds to the currently pending cube of the first i literals on the trail.
static int isRemoteExiting;
static ccc_Stats stats;
static int levelByIdCapacity;
static int* levelById;
static unsigned int progress;
int discrepancies;
double factor_th = 1000.;
#if CCC_LOG_CUBES
static FILE* cubesOutputFile = NULL;
#endif
#if CCC_LOG_SLOW_CUBES
static struct timespec* startTimeByLevel;
static FILE* slowCubesLogFile = NULL;
#endif
static int* nr_assigned_trail;
static int* part_free_trail;
FILE* cutoffCubeFile;
FILE* cutoffLearntFile;

static uint_fast64_t max_cdcl_leaf_discrepancies = 0;
static uint_fast64_t sum_cdcl_leaf_discrepancies = 0;
static uint_fast64_t max_la_leaf_discrepancies = 0;
static uint_fast64_t sum_la_leaf_discrepancies = 0;


static void printStats()
{
	extern int maxLeftBranches;
	printf("\n");
	printf("c ccc Number of cubes dispatched to CDCL:                %ld\n", stats.numberOfCubesDispatchedToCdcl);
	printf("c ccc Number of cubes proved UNSAT by CDCL:              %ld\n", stats.numberOfCubesProvedUnsatByCdcl);
	printf("c ccc Number of cubes proved UNSAT by CDCL and ignored:  %ld\n", stats.numberOfCubesProvedUnsatByCdclAndIgnored);
	printf("c ccc Number of cubes opened:                            %ld\n", stats.numberOfCubesOpened);
	printf("c ccc Number of cutoff cubes opened:                     %ld\n", stats.numberOfCutoffCubesOpened);
	printf("c ccc Number of internal cubes closed:                   %ld\n", stats.numberOfInternalCubesClosed);
	printf("c ccc Number of leaf cubes closed due to CDCL:           %ld\n", stats.numberOfLeafCubesClosedDueToCdcl);
	printf("c ccc Number of leaf cubes closed due to lookahead:      %ld\n", stats.numberOfLeafCubesClosedDueToLookahead);
	printf("c ccc Number of learnt clauses:                          %ld\n", stats.numberOfLearntClauses);
	printf("c ccc Average length of dispatched cubes:                %.3f\n", (double)stats.sumOfLengthOfDispatchedCubes / (double)stats.numberOfCubesDispatchedToCdcl);
	printf("c ccc Average length of leaf cubes:                      %.3f\n", (double)stats.sumOfLengthOfClosedLeafCubes / (double)(stats.numberOfLeafCubesClosedDueToCdcl + stats.numberOfLeafCubesClosedDueToLookahead));
	printf("c ccc Estimated progress:                                %.6f\n", (double)(progress & -(1 << (33 - nextLevel))) / 4294967296.);
	printf("c ccc Max discrepancies:                                 %d\n", max_discrepancies);
	printf("c ccc Avg discrepancies:                                 %f\n", (double)sum_discrepancies / (double)num_discrepancies);
	printf("c ccc Max left branches:                                 %d\n", maxLeftBranches);
	printf("c ccc Max number of levels backtracked:                  %d\n", stats.maxNumberOfLevelsBacktracked);
	printf("c ccc Avg number of levels backtracked:                  %.3f\n", (double)stats.sumOfNumberOfLevelsBacktracked / (double)(stats.numberOfCubesProvedUnsatByCdcl - stats.numberOfCubesProvedUnsatByCdclAndIgnored));
	printf("c ccc Final cutoff:                                      %.4f * %i\n", factor_th, part_free);
}

static void sendMessage(ccc_Message* message)
{
	// TODO: Code to check if queue is full.
	__attribute__((unused)) int result = msgsnd(queueToMinisat, message, sizeof(ccc_Message), 0);
	assert(result != -1);
}

void march_interrupt_solver()
{
#if CCC_LOG_CUBES
	fclose(cubesOutputFile);
#endif
	printStats();
}

static uint_fast64_t timeDifference(struct timespec* a, struct timespec* b)
{
	return (uint_fast64_t)(b->tv_sec - a->tv_sec) * 1000000000 + (uint_fast64_t)(b->tv_nsec - a->tv_nsec);
}

void ccc_init()
{
	nextLevel = 1;
	nextNodeId = 1;
	trail = NULL;
	isRemoteExiting = 0;
	memset(&stats, 0, sizeof(stats));
	++stats.numberOfCubesOpened; // The root node does not have a corresponding decision so it will not be opened the normal way, but it should be included in the stats to match the closed node count.
	levelByIdCapacity = 10000;
	levelById = (int*)malloc(sizeof(int) * levelByIdCapacity);
	progress = ~0;
#if CCC_LOG_CUBES
	cubesOutputFile = fopen("/tmp/cubes.txt", "w");
	if (!cubesOutputFile) { printf("Could not open 'tmp/cubes.txt'. Does the directory exist?"); exit(1); }
#endif
#if CCC_LOG_SLOW_CUBES
	startTimeByLevel = NULL;
	slowCubesLogFile = fopen("/tmp/slowCubes.txt", "w");
	if (!slowCubesLogFile) { printf("Could not open 'tmp/slowCubes.txt'. Does the directory exist?"); exit(1); }
#endif
	nr_assigned_trail = NULL;
	part_free_trail = NULL;

	if (ccc_cutoff)
	{
		cutoffCubeFile = fopen("/tmp/cubes", "w");
		if (!cutoffCubeFile) { printf("Could not open 'tmp/cubes'. Does the directory exist?"); exit(1); }
		cutoffLearntFile = fopen("/tmp/learnt", "w");
		if (!cutoffLearntFile) { printf("Could not open 'tmp/learnt'. Does the directory exist?"); exit(1); }
	}

	queueToMinisatKey = ftok(".", 0);
	if (queueToMinisatKey == -1) {
		perror("ftok");
		exit(1);
	}

	queueToMarchKey = ftok(".", 1);
	if (queueToMarchKey == -1) {
		perror("ftok");
		exit(1);
	}

	queueToMinisat = msgget(queueToMinisatKey, 0600);
	if (queueToMinisat == -1) {
		perror("msgget");
		exit(1);
	}

	queueToMarch = msgget(queueToMarchKey, 0600);
	if (queueToMarch == -1) {
		perror("msgget");
		exit(1);
	}
}

void ccc_init_trail(int variableCount)
{
#if CCC_LOG_SLOW_CUBES
	startTimeByLevel = (struct timespec*)malloc(sizeof(struct timespec) * (variableCount + 1));
	clock_gettime(CLOCK_REALTIME, &startTimeByLevel[0]);
#endif
	nr_assigned_trail = (int*)malloc(sizeof(int) * (variableCount + 1));
	part_free_trail = (int*)malloc(sizeof(int) * (variableCount + 1));
	trail = (ccc_Cube*)malloc(sizeof(ccc_Cube) * (variableCount + 1));
	trail[0].id = nextNodeId++;
	trail[0].discrepancies = 0;
	trail[0].score = 0;
	trail[0].isLeaf = 1;
	trail[0].isDiscrepancy = 0;
	trail[0].isSolvedByCdcl = 0;
	trail[0].cutoffChildCount = 0;
	trail[0].refutedChildDecision = 0;
	assert(levelByIdCapacity > 0);
	levelById[0] = 0;
}

void ccc_shutdown()
{
	printf("c march shutting down.\n");

	ccc_Message message = {1, CCC_MESSAGE_TYPE_EXIT};
	sendMessage(&message);

	// Wait for the remote end to finish so that the receive buffer will not become full and causes the remote end to block.
	while (!isRemoteExiting)
	{
		ccc_Message message;
		int receiveResult = msgrcv(queueToMarch, (void*)&message, sizeof(ccc_Message), 1, 0);
		if (receiveResult == -1)
		{
			perror("c error in march ccc_shutdown");
			break;
		}
		if (message.type == CCC_MESSAGE_TYPE_EXIT)
		{
			isRemoteExiting = 1;
		}
	}

	free(trail);
	free(levelById);
#if CCC_LOG_CUBES
	fclose(cubesOutputFile);
#endif
#if CCC_LOG_SLOW_CUBES
	free(startTimeByLevel);
	fclose(slowCubesLogFile);
#endif
	free(nr_assigned_trail);
	free(part_free_trail);
	if (ccc_cutoff)
	{
		fclose(cutoffCubeFile);
		fclose(cutoffLearntFile);
	}

	printStats();
}

void ccc_serviceMessages(int waitCubeId)
{
	while (!isRemoteExiting)
	{
		ccc_Message message;
		int receiveResult = msgrcv(queueToMarch, (void*)&message, sizeof(ccc_Message), 0, waitCubeId ? 0 : IPC_NOWAIT);
		if (receiveResult == -1)
		{
			if (errno != ENOMSG)
				perror("c error in march ccc_serviceMessages");
			break;
		}
		assert(receiveResult == sizeof(ccc_Message));

		if (message.type == CCC_MESSAGE_TYPE_UNSAT)
		{
			int32_t nodeId, level;
			++stats.numberOfCubesProvedUnsatByCdcl;
			nodeId = message.arg1;
			assert(nodeId >= 0 && nodeId < nextNodeId);
			level = levelById[nodeId];
			if (nodeId >= trail[level].id && // If the id is less, this literal was changed after having sent this cube to MiniSAT.
				level <= nextLevel - 1) // If the level is higher, march already found UNSAT, and the result by MiniSAT is of no use.
			{
				int cubeFound = 0;
				int i;
#if CCC_LOG_CUBES
				struct timespec ts;
				clock_gettime(CLOCK_REALTIME, &ts);
				fprintf(cubesOutputFile, "%ld.%09ld a ", ts.tv_sec, ts.tv_nsec);
				for (i = 1; i <= level; ++i)
				{
					fprintf(cubesOutputFile, "%d ", trail[i].decision);
				}
				fprintf(cubesOutputFile, "0\n");
#endif
				for (i = level; i < nextLevel; ++i)
				{
					if (trail[i].id == waitCubeId)
						cubeFound = 1;
					trail[i].id = 0;
					trail[i].isSolvedByCdcl = 1;
				}

				stats.sumOfNumberOfLevelsBacktracked += nextLevel - level;
				if (stats.maxNumberOfLevelsBacktracked < nextLevel - level)
					stats.maxNumberOfLevelsBacktracked = nextLevel - level;

#if SHOW_DEBUG_OUTPUT
				printf("ack. minisat proved UNSAT [% 4d]: ", nodeId);
				for (i = 1; i <= level; ++i)
					printf("%d ", trail[i].decision);
				printf("0\n");
#endif
				if (cubeFound)
					break;
			}
			else
			{
				++stats.numberOfCubesProvedUnsatByCdclAndIgnored;
#if SHOW_DEBUG_OUTPUT
				int i;
				// Ignore message because we already know it is UNSAT (note that the trail may have changed already, it is no longer accurate.)
				printf("ign. minisat proved UNSAT [% 4d]: ", nodeId);
				for (i = 1; i <= level; ++i)
					if (trail[i].id <= nodeId && i <= nextLevel - 1)
						printf("%d ", trail[i].decision);
					else
						printf("? ");
				printf("0\n");
#endif
			}
		}
		else if (message.type == CCC_MESSAGE_TYPE_EXIT)
		{
			isRemoteExiting = 1;
		}
		else
			assert(0);
	}
}

void ccc_handleDecision(int literal, float score)
{
	++stats.numberOfCubesOpened;
	if (nextLevel <= 16)
	{
		progress ^= 1 << (32 - nextLevel);
	}

#if CCC_LOG_SLOW_CUBES
	clock_gettime(CLOCK_REALTIME, &startTimeByLevel[nextLevel]);
#endif
	trail[nextLevel-1].isLeaf = 0;
	trail[nextLevel].score = score;
	trail[nextLevel].discrepancies = discrepancies;
	trail[nextLevel].id = nextNodeId;
	trail[nextLevel].decision = literal;
	trail[nextLevel].isLeaf = (trail[nextLevel-1].id != 0);
	trail[nextLevel].isDiscrepancy = trail[nextLevel].discrepancies - trail[nextLevel-1].discrepancies;
	trail[nextLevel].isSolvedByCdcl = trail[nextLevel-1].isSolvedByCdcl;
	trail[nextLevel].cutoffChildCount = 0;
	trail[nextLevel].refutedChildDecision = 0;
	trail[nextLevel].numberOfLeafNodesRefutedByCdcl = 0;
	trail[nextLevel].numberOfLeafNodesRefutedByLookahead = 0;
	if (nextNodeId >= levelByIdCapacity)
	{
		levelByIdCapacity = nextNodeId * 2;
		levelById = (int*)realloc(levelById, sizeof(int) * levelByIdCapacity);
	}
	levelById[nextNodeId] = nextLevel;

	if (trail[nextLevel-1].id == 0)
	{
#if SHOW_DEBUG_OUTPUT
		int i;
		printf("march quick skipping cube [% 4d]: ", nextNodeId);
		for (i = 1; i < nextLevel; ++i)
			printf("%d ", trail[i].decision);
		printf("%d 0\n", literal);
#endif

		trail[nextLevel].id = 0; // If the parent is UNSAT, this node is also UNSAT.
	}
	else
	{
		++stats.numberOfCubesDispatchedToCdcl;
		stats.sumOfLengthOfDispatchedCubes += nextLevel;
#if SHOW_DEBUG_OUTPUT
		int i;
		printf("march dispatching cube    [% 4d]: ", nextNodeId);
		for (i = 1; i < nextLevel; ++i)
			printf("%d ", trail[i].decision);
		printf("%d 0\n", literal);
#endif

		ccc_Message message = {1, CCC_MESSAGE_TYPE_DECISION, nextNodeId, nextLevel, literal};
		sendMessage(&message);
	}

	++nextNodeId;
	++nextLevel;
}

void ccc_handleClose(int x)
{
	int wasCutOff = x == 5;

	if (ccc_cutoff)
	{
		if (wasCutOff || trail[nextLevel-1].cutoffChildCount != 0)
		{
			// This branch was cut off, either directly or somewhere down the tree.
			if (nextLevel > 1)
				++trail[nextLevel-2].cutoffChildCount;
			if (trail[nextLevel-1].refutedChildDecision != 0)
			{
				int i;
				for (i = 1; i < nextLevel; ++i)
					fprintf(cutoffLearntFile, "%d ", -trail[i].decision);
				fprintf(cutoffLearntFile, "%d 0\n", -trail[nextLevel-1].refutedChildDecision);
				++stats.numberOfLearntClauses;
			}
		}
		else
		{
			// This cube was not cut off, nor any of its children. It really is UNSAT.
			if (nextLevel > 1)
				trail[nextLevel-2].refutedChildDecision = trail[nextLevel-1].decision;
			else
			{
				fprintf(cutoffLearntFile, "0\n"); // A bit redundant, since the conquer solver should never run in this case. But if it would, this makes it end quick.
				++stats.numberOfLearntClauses;
			}
		}
	}

	if (trail[nextLevel-1].isLeaf && !wasCutOff)
	{
#if 0
		if (trail[nextLevel - 1].isSolvedByCdcl)
			factor_th *= 1.;
		else
			factor_th *= .7;
#else
		if (trail[nextLevel - 1].isSolvedByCdcl)
		{
			double frac = .4;//depth == 0 ? 0 : 1. * (1 - discrepancies / depth);
			double newTh = .4 * (double)nr_assigned * (double)depth * (double)depth / (double)part_free;
			factor_th = factor_th * (1. - frac) + newTh * frac;
		}
		else
		{
			double frac = .4;//depth == 0 ? 0 : 1. * (1 - discrepancies / depth);
			double newTh = 3. * (double)nr_assigned * (double)depth * (double)depth / (double)part_free;
			factor_th = factor_th * (1. - frac) + newTh * frac;
		}
#endif
	}
	if (trail[nextLevel-1].isLeaf && wasCutOff)
	{
		float frac = .01;//depth == 0 ? 0 : 1. * (1 - discrepancies / depth);
		float newTh = 1.1 * (double)nr_assigned * (double)depth * (double)depth / (double)part_free;
		factor_th = factor_th * (1. - frac) + newTh * frac;
	}

	if (!wasCutOff)
	{
	if (!trail[nextLevel-1].isLeaf)
		++stats.numberOfInternalCubesClosed;
	else
	{
		stats.sumOfLengthOfClosedLeafCubes += nextLevel;
		if (trail[nextLevel - 1].isSolvedByCdcl)
		{
			int i;
			++stats.numberOfLeafCubesClosedDueToCdcl;
			for (i = 0; i < nextLevel; ++i)
				++trail[i].numberOfLeafNodesRefutedByCdcl;
			if (discrepancies > max_cdcl_leaf_discrepancies) max_cdcl_leaf_discrepancies = discrepancies;
			sum_cdcl_leaf_discrepancies += discrepancies;
		}
		else
		{
			int i;
			++stats.numberOfLeafCubesClosedDueToLookahead;
			for (i = 0; i < nextLevel; ++i)
				++trail[i].numberOfLeafNodesRefutedByLookahead;
			if (discrepancies > max_la_leaf_discrepancies) max_la_leaf_discrepancies = discrepancies;
			sum_la_leaf_discrepancies += discrepancies;
		}

		{
			clock_t clockVal = clock();
			static clock_t thresholdClockVal = 0;
			clock_t offset = clockVal - thresholdClockVal;
			if (offset >= 0)
			{
				printf("stats\t%f\t%ld\t%ld\t%ld\t%ld\t%ld\t%ld\t%ld\t%ld\t%ld\t%ld\n",
						(double)clockVal / (double)CLOCKS_PER_SEC,
						stats.numberOfCubesDispatchedToCdcl,
						sum_discrepancies,
						max_discrepancies,
						num_discrepancies,
						stats.numberOfLeafCubesClosedDueToCdcl,
						max_cdcl_leaf_discrepancies,
						sum_cdcl_leaf_discrepancies,
						stats.numberOfLeafCubesClosedDueToLookahead,
						max_la_leaf_discrepancies,
						sum_la_leaf_discrepancies);
				thresholdClockVal = thresholdClockVal + offset - offset % CLOCKS_PER_SEC + CLOCKS_PER_SEC;
			}
		}
	}
	}

#if SHOW_DEBUG_OUTPUT
	int i;
	printf("march proved/ACKed UNSAT%d [% 4d]: ", x, trail[nextLevel - 1].id);
	for (i = 1; i < nextLevel; ++i)
		printf("%d ", trail[i].decision);
	printf("0\n");
#endif

#if CCC_LOG_SLOW_CUBES
	struct timespec endTime, *startTime = &startTimeByLevel[nextLevel - 1];
	clock_gettime(CLOCK_REALTIME, &endTime);
	uint_fast64_t dt = timeDifference(startTime, &endTime);
	if (dt >= CCC_LOG_SLOW_CUBES_THRESHOLD)
	{
		int i;
		fprintf(slowCubesLogFile, "%lf\t%d\t%d\t%d\t%e\t: ", dt / 1.e9, nr_assigned_trail[nextLevel-1], part_free_trail[nextLevel-1], trail[nextLevel].discrepancies, trail[nextLevel].score);
		for (i = 1; i < nextLevel; ++i)
			fprintf(slowCubesLogFile, "%d ", trail[i].decision);
		fprintf(slowCubesLogFile, "0\n");
	}
#endif

#if 0
	if (trail[nextLevel-1].isLeaf)
	{
		int i;
		printf("\rleaf   % 8d @ %e: ", trail[nextLevel-1].id, trail[nextLevel-1].score);
		for (i = 1; i < nextLevel; ++i)
			printf("%d ", trail[i].decision);
			//printf("%c", trail[i].isDiscrepancy ? '-' : '+');
		printf("0\n");
	}
#endif

	assert(nextLevel >= 1);
	--nextLevel;
}

void ccc_handleSat()
{
#if SHOW_DEBUG_OUTPUT
	int i;
	printf("march proved SAT          [% 4d]: ", trail[nextLevel - 1].id);
	for (i = 1; i < nextLevel; ++i)
		printf("%d ", trail[i].decision);
	printf("0\n");
#endif
}

int ccc_isCurrentNodeUnsat()
{
#ifdef SIMULATE_CC
	int cutoffTriggered;
#endif
#if SHOW_DEBUG_OUTPUT
	if (trail[nextLevel - 1].id != 0)
	{
		int i;
		printf("march lookahead on cube   [% 4d]: ", trail[nextLevel - 1].id);
		for (i = 1; i < nextLevel; ++i)
			printf("%d ", trail[i].decision);
		printf("0\n");
	}
#endif
	nr_assigned_trail[nextLevel-1] = nr_assigned;
	part_free_trail[nextLevel-1] = part_free;

#ifdef SIMULATE_CC
	This is currently broken, the cutoff heuristic is computed elsewhere.
	if (cutoffTriggered)
	{
		/*{
			int i;
			printf("\rcutoff % 8d @ %e: ", trail[nextLevel-1].id, trail[nextLevel-1].score);
			for (i = 1; i < nextLevel; ++i)
				printf("%d ", trail[i].decision);
				//printf("%c", trail[i].isDiscrepancy ? '-' : '+');
			printf("0\n");
		}*/
		++stats.numberOfCutoffCubesOpened;
		if (trail[nextLevel-1].id != 0)
		{
			ccc_serviceMessages(trail[nextLevel-1].id);
		}
	}

	assert(nextLevel - 1 >= 0);
	return /*cutoffTriggered &&*/ trail[nextLevel - 1].id == 0;
#else
	return trail[nextLevel - 1].id == 0;
#endif
}

int ccc_getIsRemoteExiting()
{
	return isRemoteExiting;
}

int ccc_checkCutOff()
{
	int cutoffTriggered;

#if 0
	{
		int i;
		for (i = 1; i < nextLevel; ++i)
		{
			if (trail[i].decision != decisions[i-1])
			{
				int j;
				printf("ASSERTION FAILED: DECISIONS DO NOT MATCH!\nccc:  ");
				for (j = 1; j < nextLevel; ++j)
					printf("%d ", trail[j].decision);
				printf("\nmain: ");
				for (j = 1; j < nextLevel; ++j)
					printf("%d ", decisions[j-1]);
				printf("\n");
				exit(1);
			}
		}
	}
#endif

	if (ccc_cutoff)
	{
		//factor_th *= 1.05;
		/*if( part_free > 100000 || original_nrofclauses < 10000 )
		{
			if( depth > (30 - log2(original_nrofclauses)) ) factor_th *= 0.7;
		}
		else*/ if( depth > ccc_cutoff_depth ) factor_th *= 0.7;

		//factor_th = 5000.;
		cutoffTriggered = depth > 10 && (double)nr_assigned * (double)depth * (double)depth >= factor_th * (double)part_free;
		//cutoffTriggered = nextLevel > 1 && trail[nextLevel-1].score < 1e10;

		{
			static int calls = 0;
			++calls;
			if (calls < 1000 || (calls & 1023) == 0)
				printf("(% 7d) % 5.0f = % 8d * % 3d^2 >= % 12.3f * % 8d (% 12e)\n", stats.numberOfCutoffCubesOpened, (double)nr_assigned * (double)depth * (double)depth / (double)part_free, nr_assigned, depth, factor_th, part_free, (double)trail[nextLevel-1].score);
		}
	}
	else
	{
		cutoffTriggered = 0;
	}

	if (cutoffTriggered)
	{
		int i;
		fprintf(cutoffCubeFile, "a ");
		for (i = 1; i < nextLevel; ++i)
			fprintf(cutoffCubeFile, "%d ", trail[i].decision);
		fprintf(cutoffCubeFile, "0\n");
		++stats.numberOfCutoffCubesOpened;
#if 0
		fprintf(cutoffCubeFile, "c % 6d % 6.0f % 3d %e] ", stats.numberOfCutoffCubesOpened, factor_th, discrepancies, trail[nextLevel-1].score);
		for (i = 1; i < nextLevel; ++i)
		{
			//printf("%c", trail[i].isDiscrepancy ? '-' : '+');
			//trail[i].numberOfLeafNodesRefutedByCdcl
			fprintf(cutoffCubeFile, "(%d %d) ", trail[i].numberOfLeafNodesRefutedByCdcl, trail[i].numberOfLeafNodesRefutedByLookahead);
		}
		fprintf(cutoffCubeFile, "#\n");
#endif
	}

	return cutoffTriggered;
}

int ccc_hasGeneratedCubes()
{
	return stats.numberOfCutoffCubesOpened != 0;
}
