/* MARCH Satisfiability Solver

   Copyright (C) 2001-2009 M.J.H. Heule, J.E. van Zwieten, and M. Dufour.
   [marijn@heule.nl, jevanzwieten@gmail.com, mark.dufour@gmail.com]

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <getopt.h>
#include <float.h>
#include <fpu_control.h>

#include "march.h"
#include "assumptions.h"
#include "common.h"
#include "equivalence.h"
#include "lookahead.h"
#include "parser.h"
#include "preselect.h"
#include "progressBar.h"
#include "resolvent.h"
//#include "rootlook.h"
#include "solver.h"
#include "memory.h"
//#include "translator.h"
#include "ccc.h"

void handleUNSAT()
{
	printf( "c main():: nodeCount: %i\n", nodeCount );
	printf( "c main():: time=%f\n", ((float)(clock()))/CLOCKS_PER_SEC );
#ifdef SATTEST
	flup = fopen("results.py","w");
	fprintf(flup,"nodes=%d\n",nodeCount);
	fprintf(flup,"lookaheads=%d\n", lookAheadCount);
	fprintf(flup,"time=%f\n", ((double)(clock()))/((double)(CLOCKS_PER_SEC)) );
	fclose(flup);
#endif
#if defined(CCC)
	printf( "s UNSATISFIABLE\n" );
#elif defined(CUBE)
	printUNSAT();
#else
	printf( "s UNSATISFIABLE\n" );
#endif
	disposeFormula();

	exit( EXIT_CODE_UNSAT);
}


#ifdef CCC
#include <pthread.h>
static int ccc_initialized = 0;
void march_on_exit(void* arg)
{
	if (ccc_initialized)
		ccc_shutdown();
}

int march_core(int argc, char** argv);
int march_main(int argc, char** argv)
{
	int result;
	pthread_cleanup_push(&march_on_exit, NULL);
	result = march_core(argc, argv);
	pthread_cleanup_pop(1);
	return result;
}
#else
int march_core(int argc, char** argv);
int main(int argc, char** argv)
{
	return march_core(argc, argv);
}
#endif

int march_core( int argc, char** argv )
{
	int i, result, exitcode;
#ifdef SATTEST
        FILE *flup;
#endif

	setlinebuf(stdout);
    fpu_control_t oldcw, newcw;
    _FPU_GETCW(oldcw); newcw = (oldcw & ~_FPU_EXTENDED & ~_FPU_SINGLE) | _FPU_DOUBLE; _FPU_SETCW(newcw); // For repeatability

	/* let's not be too optimistic... :) */
	result   = UNKNOWN;
	exitcode = EXIT_CODE_UNKNOWN;

	// Parse options
	{
		const char* const USAGE = "Usage: %s [-v] [-f] [-c depth] instance_filename\n";
		int opt;
		eval_var = 0;
		full_lookahead = 0;
		ccc_cutoff = 0;
		while ((opt = getopt(argc, argv, "vfc:")) != -1) {
			switch (opt) {
			case 'v':
				eval_var = 1;
				break;
			case 'f':
				full_lookahead = 1;
				break;
			case 'c':
				if (*optarg == '=') ++optarg;
				ccc_cutoff_depth = atoi(optarg);
				ccc_cutoff = 1;
				break;
			default: /* '?' */
				fprintf(stderr, USAGE, argv[0]);
				return EXIT_CODE_ERROR;
			}
		}

		if (argc - optind != 1) {
			fprintf(stderr, USAGE, argv[0]);
			return EXIT_CODE_ERROR;
		}
	}
#ifdef PARALLEL
	if( argc - optind == 3 )
	{
	    para_depth = atoi( argv[optind+1] );
	    para_bin   = atoi( argv[optind+2] );
	}
	else
	{
	    para_depth = 0;
	    para_bin   = 0;
	}
	printf("c running in parallel using %i processors with currently number %i\n", 1 << para_depth, para_bin );
#endif
	printf("c running with eval_var %s\n", eval_var ? "enabled" : "disabled");
	printf("c running with full_lookahead %s\n", full_lookahead ? "enabled" : "disabled");
	printf("c running with ccc_cutoff %s, ccc_cutoff_depth = %d\n", ccc_cutoff ? "enabled" : "disabled", ccc_cutoff_depth);
#ifdef CCC
	ccc_init();
	ccc_initialized = 1;
#endif

	/*
		Parsing...
	*/
	runParser( argv[optind+0] );
	/*
		Preprocessing...
	*/

#ifdef PRINT_FORMULA
	compactCNF();
	printFormula( Cv );
	exit(0);
#endif
	if( equivalence_reasoning() == UNSAT )	handleUNSAT();

        for( i = 0; i < nrofclauses; i++ )
            if( Clength[ i ] > 3 )
	    { kSAT_flag = 1; break; }

	if( kSAT_flag )	printf("c using k-SAT heuristics (size based diff)\n");
	else		printf("c using 3-SAT heuristics (occurence based diff)\n");
	//check_integrety();

#ifndef TERNARYLOOK
#ifdef RESOLVENTLOOK
	if( resolvent_look() == UNSAT ) handleUNSAT();
#endif
#endif
        if( kSAT_flag )         allocate_big_clauses_datastructures();

//	TransformFormula();
//	printf( "c main():: clause / variable ratio: ( %i / %i ) = %.2f\n", nrofclauses, nrofvars, (double) nrofclauses / nrofvars );

	depth                 = 0;   // naar solver.c ?
        nodeCount             = 0;
        lookAheadCount        = 0;
        unitResolveCount      = 0;
	necessary_assignments = 0;

//	dispose_preprocessor_eq();

	if( initSolver() )			//dit staat wel heel gek...
	{
#ifdef TIMEOUT
		printf( "c timeout = %i seconds\n", TIMEOUT );
#endif
#ifdef PROGRESS_BAR
		pb_init( 6 );
#endif
#ifdef CCC
		ccc_init_trail(nrofvars);
#endif
#ifdef DISTRIBUTION
		result = distribution_branching();
#else  
#ifdef SUPER_LINEAR
		result = super_linear_branching();
#else
		result = march_solve_rec(FLT_MAX);
#endif
#endif

#ifdef PROGRESS_BAR
		pb_dispose();
#endif
	}
	else
	{
		printf( "c main():: conflict caused by unary equivalence clause found.\n" );
		result = UNSAT;
	}
#ifdef CUBE
//	if( result != UNKNOWN )
//	printDecisionTree();

        printf( "c nodes vs cubes = %i vs %i\n", nodeCount + 1, nr_cubes );
//	printf( "c ratio assigned : freevars = %i %i\n", assigned_th, part_free ); 
	printf( "c final cutoff: %.4f * %i\n", factor_th, part_free ); 
#endif
//	printf( "c main():: dead ends in main: %i\n", mainDead );
//	printf( "c main():: lookAheadCount: %i\n", lookAheadCount );
//      printf( "c main():: unitResolveCount: %i\n", unitResolveCount );
        printf( "c time = %.2f seconds\n", ((float)(clock()))/CLOCKS_PER_SEC );
//	printf( "c main():: necessary_assignments: %i\n", necessary_assignments );
//	printf( "c main():: bin_sat: %i, bin_unsat %i\n", bin_sat, bin_unsat );
#ifdef DOUBLELOOK
//	printf( "c main():: doublelook: #: %d, succes #: %d\n", (int) doublelook_count, (int) (doublelook_count - doublelook_failed) );
//	printf( "c main():: doublelook: overall %.3f of all possible doublelooks executed\n", 
//		100.0 * dl_actual_counter / (double) dl_possibility_counter );
// 	printf( "c main():: doublelook: succesrate: %.3f, average DL_trigger: %.3f\n", 100.0 - 
//		(100.0 * doublelook_failed / doublelook_count), DL_trigger_sum / doublelook_count );
#endif
#ifdef LONG_LOOK
	printf("c marin():: longlook conflicts #: %i\n", ll_conflicts );
#endif
#ifdef COUNT_SAT 
	printf( "c main():: found %i solutions\n", count_sat );
	if( count_sat > 0 ) result = SAT;
#endif

	switch( result )
	{
		case SAT:
			printf( "c main():: SOLUTION VERIFIED :-)\n" );
			printf( "s SATISFIABLE\n" );
#ifndef COUNT_SAT
		printSolution( original_nrofvars );
#endif
		exitcode = EXIT_CODE_SAT;
		break;

	    case UNSAT:
#if defined(CCC)
			if (ccc_hasGeneratedCubes())
			{
				printf("c CUTOFF\n");
				exitcode = EXIT_CODE_CUTOFF;
			}
			else
			{
				printf("s UNSATISFIABLE\n");
				exitcode = EXIT_CODE_UNSAT;
			}
#elif defined(CUBE)
			printDecisionTree();
#else
			printf( "s UNSATISFIABLE\n" );
			exitcode = EXIT_CODE_UNSAT;
#endif
		break;

	    default:
		printf( "s UNKNOWN\n" );
		exitcode = EXIT_CODE_UNKNOWN;
        }

	disposeSolver();

	disposeFormula();

        return exitcode;
}

void runParser( char* fname )
{
	FILE* in;

	if( ( in = fopen( fname, "r" ) ) == NULL )
	{
		printf( "c runParser():: input file could not be opened!\n" );
		exit( EXIT_CODE_ERROR );
	}

	if( !initFormula( in ) )
	{
		printf( "c runParser():: p-line not found in input, but required by DIMACS format!\n" );
		fclose( in );
		exit( EXIT_CODE_ERROR );
	}
	
	if( !parseCNF( in ) )
        {
                printf( "c runParser():: parse error in input!\n" );
		fclose( in );
#ifdef CCC
		pthread_exit((void*)EXIT_CODE_ERROR);
#else
		exit(EXIT_CODE_ERROR);
#endif
        }

	fclose( in );

	init_equivalence();

	if( simplify_formula() == UNSAT )
	{
		printf( "c runParser():: conflicting unary clauses, so instance is unsatisfiable!\n" );
        	printf( "s UNSATISFIABLE\n" );

		disposeFormula();
		exit( EXIT_CODE_UNSAT );
	}
}
