##
##  Template makefile for Standard, Profile, Debug, Release, and Release-static versions
##
##    eg: "make rs" for a statically linked release version.
##        "make d"  for a debug version (no optimizations).
##        "make"    for the standard version (optimized, but with debug information and assertions active)

PWD        = $(shell pwd)
EXEC      ?= $(notdir $(PWD))
OBJDIR     = $(SRCROOT)/build

CSRCS      = $(wildcard $(PWD)/*.c*)
#DSRCS      = $(foreach dir, $(DEPDIR), $(filter-out $(SRCROOT)/$(dir)/Main.cc, $(wildcard $(SRCROOT)/$(dir)/*.c*)))
DSRCS      = $(foreach dir, $(DEPDIR), $(wildcard $(SRCROOT)/$(dir)/*.c*))
CHDRS      = $(wildcard $(PWD)/*.h)
COBJS3      = $(CSRCS) $(DSRCS)
COBJS2      = $(COBJS3:.c=.c.o)
COBJS1      = $(COBJS2:.cc=.cc.o)
COBJS       = $(foreach f,$(COBJS1),$(OBJDIR)/$(subst $(SRCROOT)/,,$f))

PCOBJS     = $(addsuffix p,  $(COBJS))
DCOBJS     = $(addsuffix d,  $(COBJS))
RCOBJS     = $(addsuffix r,  $(COBJS))


CC         = gcc
CXX       ?= g++
CFLAGS    ?= -Wall -Wno-parentheses
LFLAGS    ?= -Wall

COPTIMIZE ?= -O3

CFLAGS    += -D __STDC_LIMIT_MACROS -D __STDC_FORMAT_MACROS
LFLAGS    += -lz

.PHONY : s p d r rs clean 

s:	$(EXEC)
p:	$(EXEC)_profile
d:	$(EXEC)_debug
r:	$(EXEC)_release
rs:	$(EXEC)_static

libs:	lib$(LIB)_standard.a
libp:	lib$(LIB)_profile.a
libd:	lib$(LIB)_debug.a
libr:	lib$(LIB)_release.a

## Compile options
%.o:			CFLAGS +=$(COPTIMIZE) -g -D DEBUG
%.op:			CFLAGS +=$(COPTIMIZE) -pg -g -D NDEBUG
%.od:			CFLAGS +=-O0 -g -D DEBUG
%.or:			CFLAGS +=$(COPTIMIZE) -g -D NDEBUG

## Link options
$(EXEC):		LFLAGS += -g
$(EXEC)_profile:	LFLAGS += -g -pg
$(EXEC)_debug:		LFLAGS += -g
#$(EXEC)_release:	LFLAGS += ...
$(EXEC)_static:		LFLAGS += --static

## Dependencies
$(EXEC):		$(COBJS)
$(EXEC)_profile:	$(PCOBJS)
$(EXEC)_debug:		$(DCOBJS)
$(EXEC)_release:	$(RCOBJS)
$(EXEC)_static:		$(RCOBJS)

lib$(LIB)_standard.a:	$(filter-out */Main.c.o,  $(COBJS))
lib$(LIB)_profile.a:	$(filter-out */Main.c.op, $(PCOBJS))
lib$(LIB)_debug.a:	$(filter-out */Main.c.od, $(DCOBJS))
lib$(LIB)_release.a:	$(filter-out */Main.c.or, $(RCOBJS))


## Build rule
$(OBJDIR)/%.c.o $(OBJDIR)/%.c.op $(OBJDIR)/%.c.od $(OBJDIR)/%.c.or:	$(SRCROOT)/%.c
	@echo Compiling: $(subst $(OBJDIR)/,,$@)
	@mkdir -p $(shell dirname $@)
	@$(CC) $(CFLAGS) -c -o $@ $<
$(OBJDIR)/%.cc.o $(OBJDIR)/%.cc.op $(OBJDIR)/%.cc.od $(OBJDIR)/%.cc.or:	$(SRCROOT)/%.cc
	@echo Compiling: $(subst $(OBJDIR)/,,$@)
	@mkdir -p $(shell dirname $@)
	@$(CXX) $(CFLAGS) -c -o $@ $<

## Linking rules (standard/profile/debug/release)
$(EXEC) $(EXEC)_profile $(EXEC)_debug $(EXEC)_release $(EXEC)_static:
	@echo Linking: "$@ ( $(foreach f,$^,$(subst $(OBJDIR)/,,$f)) )"
	@$(CXX) $^ $(LFLAGS) -o $@

## Library rules (standard/profile/debug/release)
lib$(LIB)_standard.a lib$(LIB)_profile.a lib$(LIB)_release.a lib$(LIB)_debug.a:
	@echo Making library: "$@ ( $(foreach f,$^,$(subst $(SRCROOT)/,,$f)) )"
	@$(AR) -rcsv $@ $^

## Library Soft Link rule:
libs libp libd libr:
	@echo "Making Soft Link: $^ -> lib$(LIB).a"
	@ln -sf $^ lib$(LIB).a

## Clean rule
clean:
	@rm -f $(EXEC) $(EXEC)_profile $(EXEC)_debug $(EXEC)_release $(EXEC)_static \
	  $(COBJS) $(PCOBJS) $(DCOBJS) $(RCOBJS) *.core $(OUTDIR)/depend.mk 

## Make dependencies
$(OBJDIR)/depend.mk: $(CSRCS) $(DSRCS) $(CHDRS)
	@echo Making dependencies
	@$(CXX) $(CFLAGS) $(CSRCS) $(DSRCS) -MM | sed 's|^\(.*\): $(SRCROOT)/\([^ ]*\)\( \\\)\?$$|$(OBJDIR)/\2.o $(OBJDIR)/\2.or $(OBJDIR)/\2.od $(OBJDIR)/\2.op: \2\3|' > $(OBJDIR)/depend.mk


-include $(OBJDIR)/depend.mk
