#include "ccc.h"
#include "mtl/Vec.h"
#include "core/SolverTypes.h"
#include "utils/Options.h"
#include <stdio.h>
#include <memory.h>
#include <assert.h>
#include <stdint.h>
#include <errno.h>
#include <sys/ipc.h>
#include <sys/msg.h>

#define SHOW_DEBUG_OUTPUT 0
#define LEARN_MARCH_UNIT_CLAUSES 0 // Warning: not compatible with -c flag!


#define CCC_MESSAGE_TYPE_EXIT 1 // {} No messages may be sent after sending a message of this type.
#define CCC_MESSAGE_TYPE_DECISION 2 // {id, level, literal} The lookahead solver makes a new cube by adding a decision literal, the cdcl solver should update its current cube upon receiving a message of this type.
#define CCC_MESSAGE_TYPE_UNSAT 3 // {id} The cdcl solver proved unsatisfiability of a cube.
typedef struct
{
	int64_t ipc_type;
	int32_t type;
	int32_t arg1;
	int32_t arg2;
	int32_t arg3;
} ccc_Message;


static key_t queueToMinisatKey = -1;
static key_t queueToMarchKey = -1;
static int queueToMinisat = -1;
static int queueToMarch = -1;
static int isRemoteExiting;
static int* idByLevel;
static int rootLevel = 0;
static bool rootRight = true;
static Minisat::SimpleBoolOption opt_eval_var       ("CCC", "v",           "Evaluate variables heuristic for march.");
static Minisat::SimpleBoolOption opt_full_lookahead ("CCC", "f",           "Full lookahead for march.");
static Minisat::IntOption opt_ccc_cutoff     ("CCC", "c",           "Act as phase-1 solver, and output cubes that were cut off.");

int ccc_getRootLevel() { return rootLevel; }


static void sendMessage(ccc_Message* message)
{
	// TODO: Code to check if queue is full.
	__attribute__((unused)) int result = msgsnd(queueToMarch, message, sizeof(ccc_Message), 0);
	assert(result != -1);
}


void ccc_init()
{
	isRemoteExiting = 0;
	idByLevel = NULL;

	queueToMinisatKey = ftok(".", 0);
	if (queueToMinisatKey == -1) {
		perror("ftok");
		exit(1);
	}

	queueToMarchKey = ftok(".", 1);
	if (queueToMarchKey == -1) {
		perror("ftok");
		exit(1);
	}

	queueToMinisat = msgget(queueToMinisatKey, 0600);
	if (queueToMinisat == -1) {
		perror("msgget");
		exit(1);
	}

	queueToMarch = msgget(queueToMarchKey, 0600);
	if (queueToMarch == -1) {
		perror("msgget");
		exit(1);
	}
}

void ccc_init_trail(int variableCount)
{
	idByLevel = (int*)malloc(sizeof(int) * (variableCount + 1));
	idByLevel[0] = 0;
}

void ccc_shutdown()
{
	printf("c minisat shutting down.\n");

	ccc_Message message = {1, CCC_MESSAGE_TYPE_EXIT};
	sendMessage(&message);

	// Wait for the remote end to finish so that the receive buffer will not become full and causes the remote end to block.
	while (!isRemoteExiting)
	{
		ccc_Message message;
		int receiveResult = msgrcv(queueToMinisat, (void*)&message, sizeof(ccc_Message), 0, 0);
		if (receiveResult == -1)
		{
			perror("c error in minisat ccc_shutdown");
			break;
		}
		if (message.type == CCC_MESSAGE_TYPE_EXIT)
		{
			isRemoteExiting = 1;
		}
	}

	free(idByLevel);
}

void ccc_serviceMessages(Minisat::vec<Minisat::Lit>& assumptions)
{
	for (;;)
	{
		ccc_Message message;
		int receiveResult = msgrcv(queueToMinisat, (void*)&message, sizeof(ccc_Message), 0, IPC_NOWAIT);
		if (receiveResult == -1)
		{
			if (errno != ENOMSG)
				perror("c error in minisat ccc_serviceMessages");
			break;
		}
		assert(receiveResult == sizeof(ccc_Message));

		if (message.type == CCC_MESSAGE_TYPE_DECISION)
		{
#if LEARN_MARCH_UNIT_CLAUSES
			if (message.arg2 == rootLevel + 1)
			{
				rootRight = !rootRight;
				//printf("\rDecision %d at root %d found (%s).\n", message.arg3, rootLevel, rootRight ? "right" : "left");
				if (rootRight) ++rootLevel;
			}
#endif
			if (message.arg2 <= assumptions.size() + 1)
			{
				idByLevel[message.arg2] = message.arg1;
				assumptions.growTo(message.arg2);
				assumptions.shrink(assumptions.size() - message.arg2); // Should be resize.
				assumptions[message.arg2 - 1].x = message.arg3 > 0 ? (message.arg3 - 1 << 1) : ((-message.arg3 - 1 << 1) + 1);
#if SHOW_DEBUG_OUTPUT
				printf("minisat checking cube     [%d]: ", message.arg1);
				for (int i = 0; i < assumptions.size(); ++i)
					printf("%d ", assumptions[i].x & 1 ? -(assumptions[i].x >> 1)-1 : (assumptions[i].x >> 1)+1);
				printf("0\n");
#endif
			}
			else
			{
#if SHOW_DEBUG_OUTPUT
				// We can only branch one literal at a time. If we are asked to branch multiple, this is because we already proved that a parent node was UNSAT. Therefore we are no longer interested in its children.
				printf("minisat discarding cube   [%d]: ", message.arg1);
				for (int i = 0; i < assumptions.size(); ++i)
					printf("%d ", assumptions[i].x & 1 ? -(assumptions[i].x >> 1)-1 : (assumptions[i].x >> 1)+1);
				for (int i = assumptions.size(); i < message.arg2 - 1; ++i)
					printf("? ");
				printf("%d 0\n", message.arg3);
#endif
			}
		}
		else if (message.type == CCC_MESSAGE_TYPE_EXIT)
		{
			isRemoteExiting = 1;
		}
		else
			assert(0);
	}
}

void ccc_handleUnsat(Minisat::vec<Minisat::Lit>& assumptions, int level)
{
#if SHOW_DEBUG_OUTPUT
	printf("minisat proved UNSAT      [%d]: ", idByLevel[level]);
	for (int i = 0; i < level; ++i)
		printf("%d ", assumptions[i].x & 1 ? -(assumptions[i].x >> 1)-1 : (assumptions[i].x >> 1)+1);
	printf("0\n");
#endif

	assert(level <= assumptions.size());

	ccc_Message message = {1, CCC_MESSAGE_TYPE_UNSAT, idByLevel[level]};
	sendMessage(&message);

	assumptions.shrinkTo(level - 1);
}

void ccc_handleSat(Minisat::vec<Minisat::Lit>& assumptions)
{
#if SHOW_DEBUG_OUTPUT
	printf("minisat proved SAT        [%d]: ", idByLevel[assumptions.size()]);
	for (int i = 0; i < assumptions.size(); ++i)
		printf("%d ", assumptions[i].x & 1 ? -(assumptions[i].x >> 1)-1 : (assumptions[i].x >> 1)+1);
	printf("0\n");
#endif
}

int ccc_updateAssumptions(Minisat::vec<Minisat::Lit>& assumptions)
{
	int previousCubeId = idByLevel[assumptions.size()];
	ccc_serviceMessages(assumptions);
	int currentCubeId = idByLevel[assumptions.size()];
	return previousCubeId != currentCubeId && !isRemoteExiting;
}

int ccc_getIsRemoteExiting()
{
	return isRemoteExiting;
}
